/*
 *
 *  Tiny Window Version 2.0
 *
 *  Copyright (c) Apr 1993
 *  Bomjun Kwon, SPARCS, KAIST (bomjun@baram.kaist.ac.kr)
 *
 */

#include "twin.h"

struct sgttyb _tty;
char _rawmode = FALSE, _pfast = FALSE, _echoit = TRUE, _reverse = FALSE;
int _res_flg, _tty_ch = 1, LINES, COLS;
char termbuf[1024], termcm[20], termcl[20], termme[20], termmr[20];
WINDOW *win[MAXWIN];
unsigned char *screen[2], *nexscr[2];
char *winmap;
char echoflag, backchar;
char halfchar = '~';

char _putchar(c)
register char c;
	{
	putchar(c);
	}

void tw_init()
	{
	register int i;
	char *tmp;

	tgetent(termbuf, getenv("TERM"));
	COLS = tgetnum("co");
	if (COLS == 0)
		COLS = 80;
	LINES = tgetnum("li");
	if (LINES == 0)
		LINES = 24;
	tmp = termcm; tgetstr("cm", &tmp);
	tmp = termcl; tgetstr("cl", &tmp);
	tmp = termme; tgetstr("me", &tmp);
	tmp = termmr; tgetstr("mr", &tmp);
	echoflag = TRUE;
	backchar = ' ';
	savetty(); cbreak(); noecho(); nonl(); noraw();
	screen[0] = (unsigned char *)malloc(LINES * COLS);
	screen[1] = (unsigned char *)malloc(LINES * COLS);
	nexscr[0] = (unsigned char *)malloc(LINES * COLS);
	nexscr[1] = (unsigned char *)malloc(LINES * COLS);
	winmap = (char *)malloc(LINES * COLS);
	for (i = 0; i < MAXWIN; i++)
		win[i] = NULL;
	
	tw_clear(); tw_normal();
	}

void tw_end()
	{
	int i;

	for (i = 0; i < MAXWIN; i++)
		{
		if (win[i] != NULL)
			{
			free(win[i]->buf[0]);
			free(win[i]->buf[1]);
			free(win[i]);
			}
		}
	free(screen[0]);
	free(screen[1]);
	free(nexscr[0]);
	free(nexscr[1]);
	free(winmap);
	tw_clear(); tw_normal();
	resetty();
	}

void tw_clear()
	{
	register int i;

	for (i = 0; i < MAXWIN; i++)
		{
		if (win[i] == NULL)
			continue;
		free(win[i]->buf[0]);
		free(win[i]->buf[1]);
		free(win[i]);
		win[i] = NULL;
		}
	for (i = 0; i < COLS * LINES; i++)
		{
		screen[0][i] = nexscr[0][i] = ' ';
		screen[1][i] = nexscr[1][i] = 0;
		winmap[i] = -1;
		}
	tputs(termcl, 0, _putchar);
	}

void tw_refresh()
	{
	register int x, y, z, w;
	register unsigned char *s0, *n0, *s1, *n1, nc;
	register char *wm, standout = FALSE;

	for (y = 0; y < LINES; y++)
		{
		s0 = screen[0] + y * COLS;
		s1 = screen[1] + y * COLS;
		n0 = nexscr[0] + y * COLS;
		n1 = nexscr[1] + y * COLS;
		wm = winmap + y * COLS;
		for (x = 0; x < COLS; x++)
			{
			nc = n0[x];
			if (n1[x] & KOREAN_1)
				{
				if (x == COLS - 1)
					nc = halfchar;
				else if (wm[x] != wm[x + 1])
					nc = halfchar;
				}
			if (n1[x] & KOREAN_2)
				{
				if (x == 0)
					nc = halfchar;
				else if (wm[x] != wm[x - 1])
					nc = halfchar;
				}
			if (s0[x] != nc || s1[x] != n1[x])
				{
				if ((n1[x] & KOREAN_2) && nc != halfchar)
					x--;
				for (z = COLS - 1; z >= 0; z--)
					{
					nc = n0[z];
					if (n1[z] & KOREAN_1)
						{
						if (z == COLS - 1)
							nc = halfchar;
						else if (wm[z] != wm[z + 1])
							nc = halfchar;
						}
					if (n1[z] & KOREAN_2)
						{
						if (z == 0)
							nc = halfchar;
						else if (wm[z] != wm[z - 1])
							nc = halfchar;
						}
					if (s0[z] != nc || s1[z] != n1[z])
						{
						if ((n1[z] & KOREAN_1) && nc != halfchar)
							z++;
						tputs(tgoto(termcm, x, y), 0, _putchar);
						for (w = x; w <= z; w++)
							{
							nc = n0[w];
							if (n1[w] & KOREAN_1)
								{
								if (w == COLS - 1)
									nc = halfchar;
								else if (wm[w] != wm[w + 1])
									nc = halfchar;
								}
							if (n1[w] & KOREAN_2)
								{
								if (w == 0)
									nc = halfchar;
								else if (wm[w] != wm[w - 1])
									nc = halfchar;
								}
							if (n1[w] & REVERSE)
								{
								if (!standout)
									{
									standout = TRUE;
									tputs(termmr, 0, _putchar);
									}
								}
							else
								{
								if (standout)
									{
									standout = FALSE;
									tputs(termme, 0, _putchar);
									}
								}
							putchar(s0[w] = nc);
							s1[w] = n1[w];
							}
						if (standout)
							{
							standout = FALSE;
							tputs(termme, 0, _putchar);
							}
						break;
						}
					}
				break;
				}
			}
		}
	fflush(stdout);
	}

WINDOW *tw_create(lx, ly, sx, sy)
register int lx, ly, sx, sy;
	{
	register int x, y, i;
	register WINDOW *w;

	for (i = 0; i < MAXWIN && win[i] != NULL; i++);
	if (i == MAXWIN)
		return NULL;
	w = (WINDOW *)malloc(sizeof(WINDOW));
	w->buf[0] = (char *)malloc(sx * sy);
	w->buf[1] = (char *)malloc(sx * sy);
	w->cx = w->cy = 0;
	w->fx = w->fy = 0;
	w->tx = sx - 1;
	w->ty = sy - 1;
	w->lx = lx; w->ly = ly; w->sx = sx; w->sy = sy;
	w->level = i;
	win[i] = w;
	for (x = 0; x < sx * sy; x++)
		{
		w->buf[0][x] = ' ';
		w->buf[1][x] = 0;
		}
	for (x = lx; x < lx + sx; x++)
		{
		for (y = ly; y < ly + sy; y++)
			{
			if (tw_properxy(x, y))
				{
				nexscr[0][x + y * COLS] = ' ';
				nexscr[1][x + y * COLS] = 0;
				winmap[x + y * COLS] = i;
				}
			}
		}
	return w;
	}

void tw_write(w, x, y, c)
register WINDOW *w;
register int x, y;
unsigned char c;
	{
	register unsigned char b0, b1;

	b0 = c;
	b1 = (_reverse) ? REVERSE : 0;
	if (c & KOREAN)
		{
		if (x == 0)
			b1 |= KOREAN_1;
		else if ((w->buf[1][x - 1 + y * w->sx] & KOREAN_1) &&
			!(w->buf[1][x + y * w->sx] & KOREAN_2))
			b1 |= KOREAN_2;
		else
			b1 |= KOREAN_1;
		}
	if (w->buf[1][x + y * w->sx] & KOREAN_1)
		{
		w->buf[0][x + 1 + y * w->sx] = ' ';
		w->buf[1][x + 1 + y * w->sx] &= REVERSE;
		if (tw_properxy(w->lx + x, w->ly + y))
			{
			if (winmap[w->lx + x + 1 + (w->ly + y) * COLS] == w->level)
				{
				nexscr[0][w->lx + x + 1 + (w->ly + y) * COLS] = ' ';
				nexscr[1][w->lx + x + 1 + (w->ly + y) * COLS] &= REVERSE;
				}
			}
		}
	if (w->buf[1][x + y * w->sx] & KOREAN_2)
		{
		w->buf[0][x - 1 + y * w->sx] = ' ';
		w->buf[1][x - 1 + y * w->sx] &= REVERSE;
		if (tw_properxy(w->lx + x, w->ly + y))
			{
			if (winmap[w->lx + x - 1 + (w->ly + y) * COLS] == w->level)
				{
				nexscr[0][w->lx + x - 1 + (w->ly + y) * COLS] = ' ';
				nexscr[1][w->lx + x - 1 + (w->ly + y) * COLS] &= REVERSE;
				}
			}
		}
	w->buf[0][x + y * w->sx] = b0;
	w->buf[1][x + y * w->sx] = b1;
	if (tw_properxy(w->lx + x, w->ly + y))
		{
		if (winmap[w->lx + x + (w->ly + y) * COLS] == w->level)
			{
			nexscr[0][w->lx + x + (w->ly + y) * COLS] = b0;
			nexscr[1][w->lx + x + (w->ly + y) * COLS] = b1;
			}
		}
	}

void tw_border(w, hc, vc, cc)
register WINDOW *w;
register char hc, vc, cc;
	{
	register int i;

	for (i = w->fx + 1; i < w->tx; i++)
		{
		tw_write(w, i, w->fy, hc);
		tw_write(w, i, w->ty, hc);
		}
	for (i = w->fy + 1; i < w->ty; i++)
		{
		tw_write(w, w->fx, i, vc);
		tw_write(w, w->tx, i, vc);
		}
	tw_write(w, w->fx, w->fy, cc);
	tw_write(w, w->fx, w->ty, cc);
	tw_write(w, w->tx, w->fy, cc);
	tw_write(w, w->tx, w->ty, cc);
	w->fx++; w->fy++; w->tx--; w->ty--;
	if (w->cx < w->fx) w->cx = w->fx;
	if (w->cx > w->tx) w->cx = w->tx;
	if (w->cy < w->fy) w->cy = w->fy;
	if (w->cy > w->ty) w->cy = w->ty;
	}

void tw_putchar(w, c)
register WINDOW *w;
register unsigned char c;
	{
	if (c == '\r')
		{
		w->cx = w->fx;
		return;
		}
	if (c == '\n')
		{
		w->cx = w->fx;
		w->cy++;
		if (w->cy > w->ty)
			{
			w->cy--;
			tw_scrollup(w);
			}
		return;
		}
	if (c == '\b')
		{
		w->cx--;
		if (w->cx < w->fx)
			{
			w->cx = w->tx;
			w->cy--;
			if (w->cy < w->fy)
				{
				w->cy = w->fy;
				tw_scrolldown(w);
				}
			}
		return;
		}
	if (c == '\t')
		{
		do
			w->cx++;
		while (w->cx % 8 == 0);
		if (w->cx > w->tx)
			tw_putchar(w, '\n');
		}
	if (c < ' ')
		return;
	if (w->cx == w->tx && (c & KOREAN))
		{
		if (!(w->buf[1][w->cx - 1 + w->cy * w->sx] & KOREAN_1) ||
			(w->buf[1][w->cx + w->cy * w->sx] & KOREAN_2))
			{
			tw_putchar(w, ' ');
			}
		}
	tw_write(w, w->cx, w->cy, c);
	w->cx++;
	if (w->cx > w->tx)
		{
		tw_putchar(w, '\n');
		}
	}

void tw_puts(w, s)
register WINDOW *w;
register char *s;
	{
	while(*s != 0)
		{
		tw_putchar(w, *s);
		s++;
		}
	}

void tw_gotoxy(w, x, y)
register WINDOW *w;
register int x, y;
	{
	w->cx = x;
	w->cy = y;
	if (w->cx < w->fx) w->cx = w->fx;
	if (w->cx > w->tx) w->cx = w->tx;
	if (w->cy < w->fy) w->cy = w->fy;
	if (w->cy > w->ty) w->cy = w->ty;
	}

void tw_erase(w)
register WINDOW *w;
	{
	register int x, y;

	w->fx = 0;
	w->fy = 0;
	w->tx = w->sx - 1;
	w->ty = w->sy - 1;
	w->cx = 0;
	w->cy = 0;
	for (y = 0; y < w->sy; y++)
		for (x = 0; x < w->sx; x++)
			tw_write(w, x, y, ' ');
	}

void tw_scrollup(w)
register WINDOW *w;
	{
	register int x, y;
	register unsigned char b0, b1;

	for (y = w->fy; y < w->ty; y++)
		for (x = w->fx; x <= w->tx; x++)
			{
			b0 = w->buf[0][x + y * w->sx] = w->buf[0][x + (y + 1) * w->sx];
			b1 = w->buf[1][x + y * w->sx] = w->buf[1][x + (y + 1) * w->sx];
			if (tw_properxy(w->lx + x, w->ly + y))
				{
				if (winmap[w->lx + x + (w->ly + y) * COLS] == w->level)
					{
					nexscr[0][w->lx + x + (w->ly + y) * COLS] = b0;
					nexscr[1][w->lx + x + (w->ly + y) * COLS] = b1;
					}
				}
			}
	for (x = w->fx; x <= w->tx; x++)
		{
		w->buf[0][x + y * w->sx] = b0 = ' ';
		w->buf[1][x + y * w->sx] = b1 = (_reverse) ? REVERSE : 0;
		if (tw_properxy(w->lx + x, w->ly + y))
			{
			if (winmap[w->lx + x + (w->ly + y) * COLS] == w->level)
				{
				nexscr[0][w->lx + x + (w->ly + y) * COLS] = b0;
				nexscr[1][w->lx + x + (w->ly + y) * COLS] = b1;
				}
			}
		}
	}

void tw_scrolldown(w)
register WINDOW *w;
	{
	register int x, y;
	register unsigned char b0, b1;

	for (y = w->ty; y > w->fy; y--)
		for (x = w->fx; x <= w->tx; x++)
			{
			b0 = w->buf[0][x + y * w->sx] = w->buf[0][x + (y - 1) * w->sx];
			b1 = w->buf[1][x + y * w->sx] = w->buf[1][x + (y - 1) * w->sx];
			if (tw_properxy(w->lx + x, w->ly + y))
				{
				if (winmap[w->lx + x + (w->ly + y) * COLS] == w->level)
					{
					nexscr[0][w->lx + x + (w->ly + y) * COLS] = b0;
					nexscr[1][w->lx + x + (w->ly + y) * COLS] = b1;
					}
				}
			}
	for (x = w->fx; x <= w->tx; x++)
		{
		w->buf[0][x + y * w->sx] = b0 = ' ';
		w->buf[1][x + y * w->sx] = b1 = (_reverse) ? REVERSE : 0;
		if (tw_properxy(w->lx + x, w->ly + y))
			{
			if (winmap[w->lx + x + (w->ly + y) * COLS] == w->level)
				{
				nexscr[0][w->lx + x + (w->ly + y) * COLS] = b0;
				nexscr[1][w->lx + x + (w->ly + y) * COLS] = b1;
				}
			}
		}
	}

void tw_front(w)
register WINDOW *w;
	{
	register int i, j, x, y;

	win[w->level] = NULL;
	for (i = 0, j = 0; i < MAXWIN; i++)
		{
		while (win[j] == NULL && j < MAXWIN) j++;
		if (j == MAXWIN)
			break;
		win[i] = win[j];
		win[i]->level = i;
		for (y = 0; y < LINES; y++)
			for (x = 0; x < COLS; x++)
				if (winmap[x + y * COLS] == j)
					winmap[x + y * COLS] = i;
		j++;
		}
	w->level = i;
	win[i++] = w;
	for (; i < MAXWIN; i++)
		win[i] = NULL;
		
	for (y = 0; y < w->sy; y++)
		for (x = 0; x < w->sx; x++)
			{
			if (!tw_properxy(w->lx + x, w->ly + y))
				continue;
			winmap[w->lx + x + (w->ly + y) * COLS] = w->level;
			nexscr[0][w->lx + x + (w->ly + y) * COLS] =
				w->buf[0][x + y * w->sx];
			nexscr[1][w->lx + x + (w->ly + y) * COLS] =
				w->buf[1][x + y * w->sx];
			}
	}

void tw_back(w)
register WINDOW *w;
	{
	register int i, j, x, y;

	for (y = 0; y < LINES; y++)
		for (x = 0; x < COLS; x++)
			if (winmap[x + y * COLS] == w->level)
				{
				for (i = w->level - 1; i >-1; i--)
					if (win[i]->lx <= x && x < win[i]->lx + win[i]->sx &&
						win[i]->ly <= y && y < win[i]->ly + win[i]->sy)
						{
						nexscr[0][x + y * COLS] = win[i]->buf[0][x - win[i]->lx
							+ (y - win[i]->ly) * win[i]->sx];
						nexscr[1][x + y * COLS] = win[i]->buf[1][x - win[i]->lx
							+ (y - win[i]->ly) * win[i]->sx];
						winmap[x + y * COLS] = i;
						break;
						}
				if (i == -1)
					{
					winmap[x + y * COLS] = -2;
					}
				}
	win[w->level] = NULL;
	for (i = 0, j = 0; i < MAXWIN; i++)
		{
		while (win[j] == NULL && j < MAXWIN) j++;
		if (j == MAXWIN)
			break;
		win[i] = win[j];
		win[i]->level = i;
		for (y = 0; y < LINES; y++)
			for (x = 0; x < COLS; x++)
				if (winmap[x + y * COLS] == j)
					winmap[x + y * COLS] = i;
		j++;
		}
	for (; i < MAXWIN; i++)
		win[i] = NULL;
	for (i = MAXWIN - 1; i > 0; i--)
		{
		if (win[i - 1] == NULL)
			continue;
		win[i] = win[i - 1];
		win[i]->level = i;
		for (y = 0; y < LINES; y++)
			for (x = 0; x < COLS; x++)
				if (winmap[x + y * COLS] == i - 1)
					winmap[x + y * COLS] = i;
		}
	win[0] = w;
	w->level = 0;
	for (y = 0; y < LINES; y++)
		for (x = 0; x < COLS; x++)
			if (winmap[x + y * COLS] == -2)
				winmap[x + y * COLS] = 0;
	}

void tw_delete(w)
register WINDOW *w;
	{
	register int i, j, x, y;

	for (y = 0; y < LINES; y++)
		for (x = 0; x < COLS; x++)
			if (winmap[x + y * COLS] == w->level)
				{
				for (i = w->level - 1; i >-1; i--)
					if (win[i]->lx <= x && x < win[i]->lx + win[i]->sx &&
						win[i]->ly <= y && y < win[i]->ly + win[i]->sy)
						{
						nexscr[0][x + y * COLS] = win[i]->buf[0][x - win[i]->lx
							+ (y - win[i]->ly) * win[i]->sx];
						nexscr[1][x + y * COLS] = win[i]->buf[1][x - win[i]->lx
							+ (y - win[i]->ly) * win[i]->sx];
						winmap[x + y * COLS] = i;
						break;
						}
				if (i == -1)
					{
					nexscr[0][x + y * COLS] = ' ';
					nexscr[1][x + y * COLS] = 0;
					winmap[x + y * COLS] = -1;
					}
				}
	win[w->level] = NULL;
	for (i = 0, j = 0; i < MAXWIN; i++)
		{
		while (win[j] == NULL && j < MAXWIN) j++;
		if (j == MAXWIN)
			break;
		win[i] = win[j];
		win[i]->level = i;
		for (y = 0; y < LINES; y++)
			for (x = 0; x < COLS; x++)
				if (winmap[x + y * COLS] == j)
					winmap[x + y * COLS] = i;
		j++;
		}
	for (; i < MAXWIN; i++)
		win[i] = NULL;

	free(w->buf[0]);
	free(w->buf[1]);
	free(w);
	}

void tw_move(w, lx, ly)
register WINDOW *w;
register int lx, ly;
	{
	register int x, y, i;

	for (y = 0; y < LINES; y++)
		for (x = 0; x < COLS; x++)
			if (winmap[x + y * COLS] == w->level)
				{
				for (i = w->level - 1; i >-1; i--)
					if (win[i]->lx <= x && x < win[i]->lx + win[i]->sx &&
						win[i]->ly <= y && y < win[i]->ly + win[i]->sy)
						{
						nexscr[0][x + y * COLS] = win[i]->buf[0][x - win[i]->lx
							+ (y - win[i]->ly) * win[i]->sx];
						nexscr[1][x + y * COLS] = win[i]->buf[1][x - win[i]->lx
							+ (y - win[i]->ly) * win[i]->sx];
						winmap[x + y * COLS] = i;
						break;
						}
				if (i == -1)
					{
					nexscr[0][x + y * COLS] = ' ';
					nexscr[1][x + y * COLS] = 0;
					winmap[x + y * COLS] = -1;
					}
				}
	w->lx = lx;
	w->ly = ly;
	for (y = 0; y < w->sy; y++)
		for (x = 0; x < w->sx; x++)
			{
			if (!tw_properxy(x + w->lx, y + w->ly))
				continue;
			if (winmap[x + w->lx + (y + w->ly) * COLS] < w->level)
				{
				nexscr[0][x + w->lx + (y + w->ly) * COLS] =
					w->buf[0][x + y * w->sx];
				nexscr[1][x + w->lx + (y + w->ly) * COLS] =
					w->buf[1][x + y * w->sx];
				winmap[x + w->lx + (y + w->ly) * COLS] = w->level;
				}
			}
	}

void tw_resize(w, sx, sy)
register WINDOW *w;
register int sx, sy;
	{
	register int x, y, i;

	for (y = 0; y < LINES; y++)
		for (x = 0; x < COLS; x++)
			if (winmap[x + y * COLS] == w->level)
				{
				for (i = w->level - 1; i >-1; i--)
					if (win[i]->lx <= x && x < win[i]->lx + win[i]->sx &&
						win[i]->ly <= y && y < win[i]->ly + win[i]->sy)
						{
						nexscr[0][x + y * COLS] = win[i]->buf[0][x - win[i]->lx
							+ (y - win[i]->ly) * win[i]->sx];
						nexscr[1][x + y * COLS] = win[i]->buf[1][x - win[i]->lx
							+ (y - win[i]->ly) * win[i]->sx];
						winmap[x + y * COLS] = i;
						break;
						}
				if (i == -1)
					{
					nexscr[0][x + y * COLS] = ' ';
					nexscr[1][x + y * COLS] = 0;
					winmap[x + y * COLS] = -1;
					}
				}
	w->buf[0] = (char *)realloc((char *)w->buf[0], sx * sy);
	w->buf[1] = (char *)realloc((char *)w->buf[1], sx * sy);
	w->cx = w->cy = 0;
	w->fx = w->fy = 0;
	w->tx = sx - 1;
	w->ty = sy - 1;
	w->sx = sx; w->sy = sy;
	for (x = 0; x < sx * sy; x++)
		{
		w->buf[0][x] = ' ';
		w->buf[1][x] = 0;
		}
	for (y = 0; y < w->sy; y++)
		for (x = 0; x < w->sx; x++)
			{
			if (!tw_properxy(x + w->lx, y + w->ly))
				continue;
			if (winmap[x + w->lx + (y + w->ly) * COLS] < w->level)
				{
				nexscr[0][x + w->lx + (y + w->ly) * COLS] = ' ';
				nexscr[1][x + w->lx + (y + w->ly) * COLS] = 0;
				winmap[x + w->lx + (y + w->ly) * COLS] = w->level;
				}
			}
	}

void tw_printf(va_alist) va_dcl
	{
	register WINDOW *w;
	register char *fmt;
	char buf[512];
	va_list pvar;

	va_start(pvar);
	w = va_arg(pvar, WINDOW *);
	fmt = va_arg(pvar, char *);
	(void) vsprintf(buf, fmt, pvar);
	va_end(pvar);
	tw_puts(w, buf);
	}

void tw_focus(w)
register WINDOW *w;
	{
	if (tw_properxy(w->cx + w->lx, w->cy + w->ly))
		tputs(tgoto(termcm, w->cx + w->lx, w->cy + w->ly), 0, _putchar);
	else
		tputs(tgoto(termcm, 0, 0), 0, _putchar);
	fflush(stdout);
	}

unsigned char tw_getchar(w)
register WINDOW *w;
	{
	register unsigned char k;
	static unsigned char han_flag = 0, han_data;

	if (han_flag == 0)
		{
		tw_refresh();
		tw_focus(w);
		}
	k = getchar();
	if (echoflag && k >= ' ')
		{
		if (han_flag == 0 && (k & KOREAN))
			{
			han_flag = 1;
			han_data = k;
			}
		else if (han_flag == 1 && (k & KOREAN))
			{
			han_flag = 0;
			tw_putchar(w, han_data);
			tw_putchar(w, k);
			}
		else
			{
			tw_putchar(w, k);
			}
		}
	return k;
	}

void tw_gets(w, s, len)
register WINDOW *w;
register char *s;
register int len;
	{
	register int i;
	register unsigned char k;
	register char tmpflag;
	static unsigned char han_flag = 0, han_data;

	tmpflag = echoflag;
	echoflag = FALSE;
	for (i = 0; i < len; i++)
		tw_putchar(w, backchar);
	for (i = 0; i < len; i++)
		tw_putchar(w, '\b');
	i = 0;

	Loop:
	k = tw_getchar(w);
	if (k == erasechar())
		{
		if (i > 0)
			{
			i--;
			tw_printf(w, "\b%c\b", backchar);
			goto Loop;
			}
		else
			goto Loop;
		}
	if (k == killchar())
		{
		while (i > 0)
			{
			tw_printf(w, "\b%c\b", backchar);
			i--;
			}
		goto Loop;
		}
	if (k == '\n' || k == '\r')
		{
		s[i] = 0;
		echoflag = tmpflag;
		for (; i < len; i++)
			tw_putchar(w, ' ');
		tw_refresh();
		return;
		}
	if (k < ' ' || i == len)
		{
		goto Loop;
		}
	s[i++] = k;
	if (tmpflag)
		{
		if (han_flag == 0 && (k & KOREAN))
			{
			han_flag = 1;
			han_data = k;
			}
		else if (han_flag == 1 && (k & KOREAN))
			{
			han_flag = 0;
			tw_putchar(w, han_data);
			tw_putchar(w, k);
			}
		else
			{
			tw_putchar(w, k);
			}
		}
	else
		tw_putchar(w, '*');
	goto Loop;
	}

void tw_redraw()
	{
	register int i;
	register char standout;

	tputs(tgoto(termcm, 0, 0), 0, _putchar);
	for (i = 0; i < LINES * COLS; i++)
		{
		if (screen[1][i] & REVERSE)
			{
			if (!standout)
				{
				standout = TRUE;
				tputs(termmr, 0, _putchar);
				}
			}
		else
			{
			if (standout)
				{
				standout = FALSE;
				tputs(termme, 0, _putchar);
				}
			}
		putchar(screen[0][i]);
		}
	}
