/*
 *
 *  Tiny Window Version 2.0
 *
 *  Copyright (c) Apr 1993
 *  Bomjun Kwon, SPARCS, KAIST (bomjun@baram.kaist.ac.kr)
 *
 */

#ifndef __TWIN
#define __TWIN

#include <stdio.h>
#include <stdlib.h>
#include <sgtty.h>
#include <varargs.h>

#define MAXWIN	(20)
#define TRUE	(1)
#define FALSE	(0)
#define ERR		(0)
#define OK		(1)

#define KOREAN		(0x80)
#define KOREAN_1	(0x01)
#define KOREAN_2	(0x02)
#define REVERSE		(0x04)

#define raw()		(_tty.sg_flags |= RAW, _pfast = _rawmode = TRUE, stty(_tty_ch, &_tty))
#define noraw()		(_tty.sg_flags &= ~RAW, _rawmode = FALSE, _pfast = !(_tty.sg_flags&CRMOD), stty(_tty_ch, &_tty))
#define cbreak()	(_tty.sg_flags |= CBREAK, _rawmode = TRUE, stty(_tty_ch, &_tty))
#define nocbreak()	(_tty.sg_flags &= ~CBREAK, _rawmode = FALSE, stty(_tty_ch, &_tty))
#define echo()		(_tty.sg_flags |= ECHO, _echoit = TRUE, stty(_tty_ch, &_tty))
#define noecho()	(_tty.sg_flags &= ~ECHO, _echoit = FALSE, stty(_tty_ch, &_tty))
#define nl()		(_tty.sg_flags |= CRMOD, _pfast = _rawmode, stty(_tty_ch, &_tty))
#define nonl()		(_tty.sg_flags &= ~CRMOD, _pfast = TRUE, stty(_tty_ch, &_tty))
#define	savetty()	((void) gtty(_tty_ch, &_tty), _res_flg = _tty.sg_flags)
#define	resetty()	(_tty.sg_flags = _res_flg, (void) stty(_tty_ch, &_tty))

#define	erasechar()	(_tty.sg_erase)
#define	killchar()	(_tty.sg_kill)
#define baudrate()	(_tty.sg_ospeed)

#define tw_normal()			(_reverse = FALSE)
#define tw_reverse()		(_reverse = TRUE)
#define tw_properxy(x, y)	((x) >=0 && (x) < COLS && (y) >=0 && (y) < LINES)
#define tw_char(w, x, y)	((w)->buf[0][(x) + (y) * (w)->sx])
#define tw_attr(w, x, y)	((w)->buf[1][(x) + (y) * (w)->sx])
#define tw_cursorx(w)		((w)->cx)
#define tw_cursory(w)		((w)->cy)
#define tw_sizex(w)			((w)->sx)
#define tw_sizey(w)			((w)->sy)
#define tw_locx(w)			((w)->lx)
#define tw_loy(w)			((w)->ly)
#define tw_fromx(w)			((w)->fx)
#define tw_fromy(w)			((w)->fy)
#define tw_tox(w)			((w)->tx)
#define tw_toy(w)			((w)->ty)
#define tw_level(w)			((w)->level)
#define tw_backchar(c)		(backchar = (c))
#define tw_showchar()		(echoflag = TRUE)
#define tw_hidechar()		(echoflag = FALSE)
#define tw_halfchar(c)		(halfchar = (c))

typedef struct
	{
	char *buf[2];						/* screen buffer */
	int cx, cy;							/* relative location of cursor */
	int fx, fy, tx, ty;					/* active area */
	int lx, ly;							/* absolute location of window */
	int sx, sy;							/* size of window */
	int level;							/* level of window */
	} WINDOW;

extern struct sgttyb _tty;
extern char _rawmode, _pfast, _echoit, _reverse, echoflag, backchar;
extern int _res_flg, _tty_ch;

extern LINES, COLS;						/* size of current screen */
extern int WINS;						/* current number of windows */
extern char halfchar;					/* character for half Korean */
extern WINDOW *win[MAXWIN];				/* pointers to windows */
extern unsigned char *screen[2];		/* current contents of screen */
extern unsigned char *nexscr[2];		/* next contents of screen */
extern char *winmap;					/* map of window region */

extern void		tw_init();
extern void		tw_end();
extern void		tw_clear();
extern WINDOW	*tw_create();
extern void		tw_refresh();
extern void		tw_border();
extern void		tw_write();
extern void		tw_putchar();
extern void		tw_puts();
extern void		tw_gotoxy();
extern void		tw_erase();
extern void		tw_scrollup();
extern void		tw_scrolldown();
extern void		tw_front();
extern void		tw_back();
extern void		tw_move();
extern void		tw_resize();
extern void		tw_delete();
extern void		tw_printf();
extern void		tw_focus();
extern void		tw_gets();
extern void		tw_redraw();
extern unsigned char tw_getchar();

#endif
